from .fields import format_fields


def test_format_fields__string():
    fields = "a"

    assert format_fields(fields) == "a"


def test_format_fields__strings():
    fields = ("a", "b", "c", "d")

    assert format_fields(fields) == "a, b, c, d"


def test_format_fields__dict():
    fields = {"a": ("1", "2")}

    assert format_fields(fields) == "a(1, 2)"


def test_format_fields__dicts():
    fields = ({"a": ("1", "2")}, {"b": ("3", "4", "5")})

    assert format_fields(fields) == "a(1, 2), b(3, 4, 5)"


def test_format_fields__nested_dict():
    fields = {"a": {"b": {"c": ("1", "2")}}}

    assert format_fields(fields) == "a(b(c(1, 2)))"


def test_format_fields__wraped_tuple():
    fields = (("1",), ({"a": (("4",),)}))

    assert format_fields(fields) == "1, a(4)"



from typing import NamedTuple, Optional

from ...utils import OUTPUT_DIR, from_env, validate_baseurl
from .constant import (
    BASE_URL,
    CLIENT_ID,
    CLIENT_SECRET,
    LOG_TO_STDOUT,
    SEARCH_PER_FOLDER,
)
from .env import thread_pool_size


def _valid_thread_pool_size(**kwargs):
    """
    Checks that the thread pool size is between 1 and 200 inclusive.
    https://www.googlecloudcommunity.com/gc/Technical-Tips-Tricks/Is-there-a-limit-to-the-number-of-API-requests-you-can-send-at/ta-p/592612
    """
    size = kwargs.get("thread_pool_size") or thread_pool_size()
    if not 1 <= size <= 200:
        raise ValueError("Thread pool size must be between 1 and 200 inclusive")
    return size


def _bool_env_variable(key: str) -> bool:
    parameter = from_env(key, allow_missing=True)
    return str(parameter).lower() == "true"


class Parameters(NamedTuple):
    """Parameters for Looker extraction"""

    base_url: str
    client_id: str
    client_secret: str
    is_safe_mode: bool
    log_to_stdout: bool
    output_directory: str
    search_per_folder: bool
    thread_pool_size: int
    timeout: Optional[int]


def get_parameters(**kwargs) -> Parameters:
    """
    Returns parameters for Looker extraction whether they come from script
    argument, env variable or default value.
    """
    output_directory = kwargs.get("output_directory") or from_env(OUTPUT_DIR)

    base_url = validate_baseurl(kwargs.get("base_url") or from_env(BASE_URL))
    client_id = kwargs.get("client_id") or from_env(CLIENT_ID)
    client_secret = kwargs.get("client_secret") or from_env(CLIENT_SECRET)

    thread_pool_size_ = _valid_thread_pool_size(**kwargs)

    # timeout can be set with env variable however we don't use from_env because it has a default value
    timeout = kwargs.get("timeout")

    is_safe_mode = kwargs.get("safe_mode", False)
    search_per_folder = kwargs.get("search_per_folder") or _bool_env_variable(
        SEARCH_PER_FOLDER,
    )
    log_to_stdout = kwargs.get("log_to_stdout") or _bool_env_variable(
        LOG_TO_STDOUT
    )

    return Parameters(
        base_url=base_url,
        client_id=client_id,
        client_secret=client_secret,
        is_safe_mode=is_safe_mode,
        log_to_stdout=log_to_stdout,
        output_directory=output_directory,
        search_per_folder=search_per_folder,
        thread_pool_size=thread_pool_size_,
        timeout=timeout,
    )



from typing import List, Optional, Tuple
from urllib.parse import urlsplit, urlunsplit

BASE_URL_SCHEME = "https"


class InvalidBaseUrl(ValueError):
    ...


def _preprocess_url(base_url: str) -> str:
    if "://" not in base_url:
        return f"{BASE_URL_SCHEME}://{base_url}"

    return base_url.strip()


def _clean_path(path: str) -> str:
    return path.rstrip("/")


def _get_hostname_port(netloc: str) -> Tuple[str, str]:
    hostname, *rest = netloc.split(":")
    port = ":".join(rest)
    return hostname, port


def _urlsplit(base_url: str) -> Tuple[str, str, str, str, str, str]:
    """
    Returns URL split parts

    Parameters:
            base_url (str): The URL used for the splitting.

    Returns:
            scheme (str)
            hostname (str)
            path (str)
            port (str)
            query (str)
            fragment (str)
    """

    base_url = _preprocess_url(base_url)
    url = urlsplit(base_url)

    hostname, port = _get_hostname_port(url.netloc)
    path = _clean_path(url.path)

    return url.scheme, hostname, path, port, url.query, url.fragment


def _expect(
    attr: str,
    expected: Optional[List[str]],
    actual: Optional[str],
) -> None:
    if not expected and not actual:
        return
    if expected and actual in expected:
        return
    raise InvalidBaseUrl(
        f"Invalid base url {attr} | expected: {expected}, got: {actual}",
    )


def validate_baseurl(base_url: str) -> str:
    scheme, hostname, path, port, query, fragment = _urlsplit(base_url)
    hostname_with_port = f"{hostname}:{port}" if port else hostname

    if not hostname:
        raise InvalidBaseUrl(f"Invalid base url hostname: {hostname}")

    _expect("scheme", ["https", "http"], scheme)
    _expect("query", None, query)
    _expect("fragment", None, fragment)
    result = urlunsplit((scheme, hostname_with_port, path, "", ""))

    return result



import ast
import re
import string
from typing import Tuple, TypeVar, Union

_ALPHANUMERIC_REGEX = r"^\w+$"
_SYMBOLS = "[]{}()"

T = TypeVar("T", bytes, str, int, float, bool, None)


def _clean(input_: str) -> str:
    """
    Clean the input to reduce the number of cases:
    - remove trailing white spaces
    - remove carriage returns '\n'
    - remove lists symbols ( ) [ ] { }
      "  [  'a', 'b'  ]  " => "'a', 'b'"
      ['a', 'b'] ~ {'a', 'b'} ~ ('a', 'b') => 'a', 'b'
      ['a'] ~ {'a'} ~ ('a') => 'a'
      a => 'a'
    """
    # remove trailing symbols and whitespaces
    cleaned = input_.strip(_SYMBOLS + string.whitespace)
    cleaned = cleaned.replace("\n", "")

    # handle unquoted singleton: _clean(foo) == 'foo'
    if re.match(_ALPHANUMERIC_REGEX, cleaned):
        cleaned = f"'{cleaned}'"

    return cleaned


def string_to_tuple(input_: str) -> Tuple[str, ...]:
    """
    Parse the given string and returns the corresponding Tuple of strings

    Supported formats:
    - list ['a', 'b']
    - tuples ('a', 'b')
    - sets {'a', 'b', 'c'}

    Extra coma allowed: ['a', 'b', ]
    Carriage returns allowed: [
        'a',
        'b'
    ]
    Empty list allowed: []
    Singleton allowed: 'a'
    Multiple types allowed: ['foo', 100, 19.8]
    Comas between double quotes allowed: "Hi, there" => ("Hi, there", ...)

    Not supported:
    - unbalanced quotes: "a", b"
    - Multiple strings without quotes: a, b
    """
    cleaned = _clean(input_)
    if not cleaned:
        # support empty string
        return tuple()

    parsed = ast.literal_eval(cleaned)

    # list or tuple containing several elements
    if isinstance(parsed, (tuple, list, set)):
        return tuple(str(element) for element in parsed)

    # float or integer singleton
    if isinstance(parsed, (int, float)):
        return (str(parsed),)

    # string singleton
    if isinstance(parsed, str):  # single element
        return (parsed,)

    raise ValueError(f"Could not parse input to Tuple[String,...]: {input_}")


def decode_when_bytes(value: T, encoding: str = "utf-8") -> Union[T, str]:
    """
    Decode bytes using the given encoding, if no encoding are given utf-8 will
    be used by default.
    If value is not bytes, then the value is returned as-is.
    """
    if isinstance(value, bytes):
        return value.decode(encoding)

    return value



import logging
from argparse import ArgumentParser

from castor_extractor.warehouse import snowflake  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    group = parser.add_mutually_exclusive_group()
    group.add_argument("-p", "--password", help="Snowflake password")
    group.add_argument("-pk", "--private-key", help="Snowflake private key")

    parser.add_argument("-a", "--account", help="Snowflake account")
    parser.add_argument("-u", "--user", help="Snowflake user")

    parser.add_argument(
        "--warehouse",
        help="Use a specific WAREHOUSE to run extraction queries",
    )
    parser.add_argument(
        "--role",
        help="Use a specific ROLE to run extraction queries",
    )

    parser.add_argument(
        "--db-allowed",
        nargs="*",
        help="List of databases that should be extracted",
    )
    parser.add_argument(
        "--db-blocked",
        nargs="*",
        help="List of databases that should not be extracted",
    )
    parser.add_argument(
        "--fetch-transient",
        action="store_true",
        help="Optional: will fetch transients tables if added",
    )

    parser.add_argument("-o", "--output", help="Directory to write to")

    parser.add_argument(
        "--skip-existing",
        dest="skip_existing",
        action="store_true",
        help="Skips files already extracted instead of replacing them",
    )
    parser.set_defaults(skip_existing=False)

    args = parser.parse_args()

    snowflake.extract_all(
        account=args.account,
        user=args.user,
        password=args.password,
        private_key=args.private_key,
        warehouse=args.warehouse,
        role=args.role,
        db_allowed=args.db_allowed,
        db_blocked=args.db_blocked,
        output_directory=args.output,
        skip_existing=args.skip_existing,
        fetch_transient=args.fetch_transient,
    )



import logging
from typing import Iterator

from psycopg2 import extensions  # type: ignore
from sqlalchemy.engine import Connection, ResultProxy

from ...utils import decode_when_bytes, uri_encode
from ..abstract import SqlalchemyClient

REDSHIFT_URI = "redshift+psycopg2://{user}:{password}@{host}:{port}/{database}"

DEFAULT_PORT = 5439

logger = logging.getLogger(__name__)


class RedshiftClient(SqlalchemyClient):
    """redshift client"""

    @staticmethod
    def name() -> str:
        return "Redshift"

    def _set_binary_connection(self) -> None:
        """
        Get the underlying psycopg2 connection and
        set the connection to BINARY mode
        """
        assert self._connection, "Connection missing"
        conn = self._connection.connection.connection
        extensions.register_type(extensions.BYTES, conn)

    def connect(self) -> Connection:
        connection = super().connect()
        self._set_binary_connection()
        return connection

    @staticmethod
    def _process_result(proxy: ResultProxy) -> Iterator[dict]:
        """
        Some clients using redshift have rows with mixed encoding, this
        function discard those problematics rows.
        """
        total_rows, decode_errors_count = 0, 0
        for row in proxy:
            total_rows += 1
            try:
                yield {
                    key: decode_when_bytes(value)
                    for key, value in dict(row).items()
                }
            except UnicodeDecodeError:
                decode_errors_count += 1
                continue

        if decode_errors_count > 0:
            logger.info(
                f"{decode_errors_count} UnicodeDecodeError on {total_rows} rows"
            )

    def _engine_options(self, credentials: dict) -> dict:
        return {"connect_args": {"sslmode": "verify-ca"}}

    def _build_uri(self, credentials: dict) -> str:
        return REDSHIFT_URI.format(
            user=credentials["user"],
            password=uri_encode(credentials["password"]),
            host=credentials["host"],
            port=credentials.get("port") or DEFAULT_PORT,
            database=credentials["database"],
        )



import logging
from datetime import date, datetime, timedelta
from time import sleep
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, Union

import msal  # type: ignore
import requests

from ....utils import at_midnight, current_date
from ..assets import PowerBiAsset
from .constants import (
    DEFAULT_TIMEOUT_IN_SECS,
    GET,
    POST,
    SCAN_READY,
    Batches,
    Keys,
    QueryParams,
    Urls,
)
from .credentials import Credentials
from .utils import batch_size_is_valid_or_assert, datetime_is_recent_or_assert

logger = logging.getLogger(__name__)

_FORMAT = "%Y-%m-%dT%H:%M:%S"


def _format(timestamp: datetime) -> str:
    return timestamp.strftime(_FORMAT)


def _yesterday() -> date:
    return current_date() - timedelta(days=1)


def _time_filter(day: Optional[date]) -> Tuple[datetime, datetime]:
    target_day = day or _yesterday()
    start = at_midnight(target_day)
    end = datetime.combine(target_day, datetime.max.time())
    return start, end


def _url(
    day: Optional[date],
    continuation_uri: Optional[str],
) -> str:
    if continuation_uri:
        return continuation_uri

    url = Urls.ACTIVITY_EVENTS
    start, end = _time_filter(day)
    url += "?$filter=Activity eq 'viewreport'"
    url += f"&startDateTime='{_format(start)}'"
    url += f"&endDateTime='{_format(end)}'"
    return url


class Client:
    """
    PowerBI rest admin api
    https://learn.microsoft.com/en-us/rest/api/power-bi/admin
    """

    def __init__(self, credentials: Credentials):
        self.creds = credentials

    def _access_token(self) -> dict:
        client_app = f"{Urls.CLIENT_APP_BASE}{self.creds.tenant_id}"
        app = msal.ConfidentialClientApplication(
            client_id=self.creds.client_id,
            authority=client_app,
            client_credential=self.creds.secret,
        )

        token = app.acquire_token_silent(self.creds.scopes, account=None)

        if not token:
            token = app.acquire_token_for_client(scopes=self.creds.scopes)

        if Keys.ACCESS_TOKEN not in token:
            raise ValueError(f"No access token in token response: {token}")

        return token

    def _header(self) -> Dict:
        """Return header used in following rest api call"""
        token = self._access_token()
        return {"Authorization": f"Bearer {token[Keys.ACCESS_TOKEN]}"}

    def _call(
        self,
        url: str,
        method: str = GET,
        *,
        params: Optional[Dict] = None,
        data: Optional[dict] = None,
        processor: Optional[Callable] = None,
    ) -> Any:
        """
        Make either a get or a post http request.Request, by default
        result.json is returned. Optionally you can provide a processor callback
        to transform the result.
        """
        logger.debug(f"Calling {method} on {url}")
        result = requests.request(
            method,
            url,
            headers=self._header(),
            params=params,
            data=data,
        )
        result.raise_for_status()

        if processor:
            return processor(result)

        return result.json()

    def _get(
        self,
        url: str,
        *,
        params: Optional[Dict] = None,
        processor: Optional[Callable] = None,
    ) -> Any:
        return self._call(url, GET, params=params, processor=processor)

    def _post(
        self,
        url: str,
        *,
        params: Optional[dict],
        data: Optional[dict],
        processor: Optional[Callable] = None,
    ) -> Any:
        return self._call(
            url,
            POST,
            params=params,
            data=data,
            processor=processor,
        )

    def _workspace_ids(
        self,
        modified_since: Optional[datetime] = None,
    ) -> List[str]:
        """
        Get workspaces ids from powerBI admin API.
        If modified_since, take only workspaces that have been modified since

        more: https://learn.microsoft.com/en-us/rest/api/power-bi/admin/workspace-info-get-modified-workspaces
        """

        def result_callback(call_result: requests.models.Response) -> List[str]:
            return [x["id"] for x in call_result.json()]

        params: Dict[str, Union[bool, str]] = {
            Keys.INACTIVE_WORKSPACES: True,
            Keys.PERSONAL_WORKSPACES: True,
        }

        if modified_since:
            datetime_is_recent_or_assert(modified_since)
            modified_since_iso = f"{modified_since.isoformat()}0Z"
            params[Keys.MODIFIED_SINCE] = modified_since_iso

        result = self._get(
            Urls.WORKSPACE_IDS,
            params=params,
            processor=result_callback,
        )

        return result

    def _create_scan(self, workspaces_ids: List[str]) -> int:
        batch_size_is_valid_or_assert(workspaces_ids)
        request_body = {"workspaces": workspaces_ids}
        params = QueryParams.METADATA_SCAN
        scan_id = self._post(
            Urls.METADATA_POST,
            params=params,
            data=request_body,
        )
        return scan_id[Keys.ID]

    def _wait_for_scan_result(self, scan_id: int) -> bool:
        url = f"{Urls.METADATA_WAIT}/{scan_id}"
        waiting_seconds = 0
        sleep_seconds = 1
        while True:
            result = self._get(url, processor=lambda x: x)
            if result.status_code != 200:
                return False
            if result.json()[Keys.STATUS] == SCAN_READY:
                logger.info(f"scan {scan_id} ready")
                return True
            if waiting_seconds >= DEFAULT_TIMEOUT_IN_SECS:
                break
            waiting_seconds += sleep_seconds
            logger.info(
                f"Waiting {sleep_seconds} sec for scan {scan_id} to be ready…",
            )
            sleep(sleep_seconds)
        return False

    def _get_scan(self, scan_id: int) -> List[dict]:
        url = f"{Urls.METADATA_GET}/{scan_id}"
        return self._get(url)[Keys.WORKSPACES]

    def _activity_events(
        self,
        *,
        day: Optional[date] = None,
        continuation_uri: Optional[str] = None,
    ) -> List[Dict]:
        """
        Returns a list of activity events for the organization.
        https://learn.microsoft.com/en-us/power-bi/admin/service-admin-auditing#activityevents-rest-api
        - when no day is specified, fallback is yesterday
        - continuation_uri allows to fetch paginated data (internal usage)
        """
        url = _url(day, continuation_uri)
        answer = self._get(url)
        activity_events = answer[Keys.ACTIVITY_EVENT_ENTITIES]
        is_last = answer[Keys.LAST_RESULT_SET]
        assert isinstance(is_last, bool)
        if is_last:
            return activity_events

        # there are more data to fetch
        # https://learn.microsoft.com/en-us/rest/api/power-bi/admin/get-activity-events#get-the-next-set-of-audit-activity-events-by-sending-the-continuation-token-to-the-api-example
        continuation_uri = answer[Keys.CONTINUATION_URI]
        rest = self._activity_events(continuation_uri=continuation_uri)
        activity_events.extend(rest)
        return activity_events

    def _datasets(self) -> List[Dict]:
        """
        Returns a list of datasets for the organization.
        https://learn.microsoft.com/en-us/rest/api/power-bi/admin/datasets-get-datasets-as-admin
        """
        return self._get(Urls.DATASETS)[Keys.VALUE]

    def _reports(self) -> List[Dict]:
        """
        Returns a list of reports for the organization.
        https://learn.microsoft.com/en-us/rest/api/power-bi/admin/reports-get-reports-as-admin
        """
        return self._get(Urls.REPORTS)[Keys.VALUE]

    def _dashboards(self) -> List[Dict]:
        """
        Returns a list of dashboards for the organization.
        https://learn.microsoft.com/en-us/rest/api/power-bi/admin/dashboards-get-dashboards-as-admin
        """
        return self._get(Urls.DASHBOARD)[Keys.VALUE]

    def _metadata(
        self,
        modified_since: Optional[datetime] = None,
    ) -> Iterator[List[Dict]]:
        """
        Fetch metadata by workspace.
        https://learn.microsoft.com/en-us/power-bi/enterprise/service-admin-metadata-scanning
        """
        ids = self._workspace_ids(modified_since)

        for ix in range(0, len(ids), Batches.METADATA):
            batch_ids = [w_id for w_id in ids[ix : ix + Batches.METADATA]]
            scan_id = self._create_scan(batch_ids)
            self._wait_for_scan_result(scan_id)
            yield self._get_scan(scan_id)

    def fetch(
        self,
        asset: PowerBiAsset,
        *,
        modified_since: Optional[datetime] = None,
        day: Optional[date] = None,
    ) -> List[Dict]:
        """
        Given a PowerBi asset, returns the corresponding data using the
        appropriate client.
        """
        logger.info(f"Starting extraction of {asset}")
        asset = PowerBiAsset(asset)

        if asset == PowerBiAsset.ACTIVITY_EVENTS:
            return self._activity_events(day=day)

        if asset == PowerBiAsset.DATASETS:
            return self._datasets()

        if asset == PowerBiAsset.DASHBOARDS:
            return self._dashboards()

        if asset == PowerBiAsset.REPORTS:
            return self._reports()

        assert asset == PowerBiAsset.METADATA
        return [
            item for batch in self._metadata(modified_since) for item in batch
        ]



import logging
from typing import Dict, Iterator, List, Optional, Tuple
from urllib.parse import urljoin

import requests

from ..assets import SigmaAsset
from .credentials import CredentialsKey, SigmaCredentials
from .endpoints import EndpointFactory
from .pagination import Pagination

logger = logging.getLogger()


DATA_ELEMENTS: Tuple[str, str] = ("table", "visualization")


class SigmaClient:
    """Client used for all Sigma's assets extractions"""

    def __init__(self, credentials: SigmaCredentials):
        self.host = credentials.host
        self.client_id = credentials.client_id
        self.api_token = credentials.api_token
        self.headers: Optional[Dict[str, str]] = None

    def _get_token(self, token_api_path: str) -> Dict[str, str]:
        token_response = requests.post(  # noqa: S113
            token_api_path,
            data={
                CredentialsKey.GRANT_TYPE.value: "client_credentials",
                CredentialsKey.CLIENT_ID.value: self.client_id,
                CredentialsKey.CLIENT_SECRET.value: self.api_token,
            },
        )
        if token_response.status_code != 200:
            raise ValueError("Couldn't fetch the token in the API")
        return token_response.json()

    def _get_headers(self) -> Dict[str, str]:
        if not self.headers:
            headers = {"Content-Type": "application/x-www-form-urlencoded"}

            auth_endpoint = EndpointFactory.authentication()
            token_api_path = urljoin(self.host, auth_endpoint)

            token = self._get_token(token_api_path)
            headers["Authorization"] = f"Bearer {token['access_token']}"
            self.headers = headers

        return self.headers

    def _get(self, endpoint_url: str) -> dict:
        url = urljoin(self.host, endpoint_url)
        result = requests.get(url, headers=self._get_headers())  # noqa: S113
        try:
            return result.json()
        except:
            logger.debug(result)
            return dict()

    def _get_with_pagination(self, endpoint_url: str) -> Iterator[dict]:
        pagination = Pagination(next_page="0")

        while pagination.next_page is not None:
            paginated_url = pagination.generate_url(endpoint_url)
            response = self._get(paginated_url)
            pagination = Pagination(
                next_page=response.get("nextPage"),
                entries=response.get("entries"),
                total=response.get("total"),
            )
            yield from pagination.entries

    def _per_workbook_get_pages(self, workbook_id: str) -> Iterator[dict]:
        endpoint = EndpointFactory.pages(workbook_id)
        yield from self._get_with_pagination(endpoint)

    def _per_page_get_elements(
        self,
        workbook_id: str,
        page_id: str,
    ) -> Iterator[dict]:
        endpoint = EndpointFactory.elements(workbook_id, page_id)
        yield from self._get_with_pagination(endpoint)

    def _get_all_datasets(self) -> Iterator[dict]:
        endpoint = EndpointFactory.datasets()
        yield from self._get_with_pagination(endpoint)

    def _get_all_files(self) -> Iterator[dict]:
        endpoint = EndpointFactory.files()
        yield from self._get_with_pagination(endpoint)

    def _get_all_members(self) -> Iterator[dict]:
        endpoint = EndpointFactory.members()
        yield from self._get(endpoint)

    def _get_all_workbooks(self) -> Iterator[dict]:
        endpoint = EndpointFactory.workbooks()
        yield from self._get_with_pagination(endpoint)

    def _get_all_elements(self, workbooks: List[dict]) -> Iterator[dict]:
        for workbook in workbooks:
            workbook_id = workbook["workbookId"]
            pages = self._per_workbook_get_pages(workbook_id)

            for page in pages:
                page_id = page["pageId"]
                elements = self._per_page_get_elements(workbook_id, page_id)
                for element in elements:
                    if element.get("type") not in DATA_ELEMENTS:
                        continue
                    yield {
                        **element,
                        "workbook_id": workbook_id,
                        "page_id": page_id,
                    }

    def _get_all_lineages(self, elements: List[dict]) -> Iterator[dict]:
        for element in elements:
            workbook_id = element["workbook_id"]
            element_id = element["elementId"]
            endpoint = EndpointFactory.lineage(workbook_id, element_id)
            lineage = self._get(endpoint)
            yield {
                **lineage,
                "workbook_id": workbook_id,
                "element_id": element_id,
            }

    def _get_all_queries(self, workbooks: List[dict]) -> Iterator[dict]:
        for workbook in workbooks:
            workbook_id = workbook["workbookId"]
            endpoint = EndpointFactory.queries(workbook_id)
            queries = self._get_with_pagination(endpoint)
            for query in queries:
                yield {**query, "workbook_id": workbook_id}

    def fetch(
        self,
        asset: SigmaAsset,
        workbooks: Optional[List[dict]] = None,
        elements: Optional[List[dict]] = None,
    ) -> Iterator[dict]:
        """Returns the needed metadata for the queried asset"""
        if asset == SigmaAsset.DATASETS:
            yield from self._get_all_datasets()

        elif asset == SigmaAsset.ELEMENTS:
            if not workbooks:
                raise ValueError("Missing workbooks to extract elements")

            yield from self._get_all_elements(workbooks)

        elif asset == SigmaAsset.FILES:
            yield from self._get_all_files()

        elif asset == SigmaAsset.LINEAGES:
            if not elements:
                raise ValueError("Missing elements to extract lineage")
            yield from self._get_all_lineages(elements)

        elif asset == SigmaAsset.MEMBERS:
            yield from self._get_all_members()

        elif asset == SigmaAsset.QUERIES:
            if not workbooks:
                raise ValueError("Missing workbooks to extract queries")

            yield from self._get_all_queries(workbooks)

        elif asset == SigmaAsset.WORKBOOKS:
            yield from self._get_all_workbooks()

        else:
            raise ValueError(f"This asset {asset} is unknown")



from typing import Optional
from unittest.mock import call, patch

from .rest import RestApiClient


def _check_called_once(
    client: RestApiClient,
    first_page_url: str,
    return_value: Optional[dict],
):
    with patch.object(RestApiClient, "_call") as mock_call:
        mock_call.return_value = return_value
        data = client._pager(first_page_url)

        expected = [] if not return_value else return_value["data"]
        assert data == expected

        mock_call.assert_called_once_with(first_page_url)


def test_rest_api_client_pager():
    dummy_server_url = "https://clic.kom"
    dummy_api_key = "i-am-the-key-dont-let-others-know-about"

    client = RestApiClient(server_url=dummy_server_url, api_key=dummy_api_key)

    first_page_url = "https://clic.kom/assets"

    # no response -> 1 call
    _check_called_once(client, first_page_url, return_value=None)

    # no next page -> 1 call
    return_value = {"data": [1], "links": {}}
    _check_called_once(client, first_page_url, return_value=return_value)

    # next page = current page -> 1 call
    return_value = {"data": [1], "links": {"next": {"href": first_page_url}}}
    _check_called_once(client, first_page_url, return_value=return_value)

    # next page -> 2 calls
    another_page_url = "https://clic.kom/assets?page2"
    return_value = {"data": [1], "links": {"next": {"href": another_page_url}}}

    with patch.object(RestApiClient, "_call") as mock_call:
        mock_call.return_value = return_value

        data = client._pager(first_page_url)
        assert data == [1, 1]

        calls = [call(first_page_url), call(another_page_url)]
        mock_call.assert_has_calls(calls)



import logging
from collections import OrderedDict

from ..abstract import (
    CATALOG_ASSETS,
    QUERIES_ASSETS,
    VIEWS_ASSETS,
    SupportedAssets,
    WarehouseAsset,
    WarehouseAssetGroup,
)

logger = logging.getLogger(__name__)

SYNAPSE_ASSETS: SupportedAssets = OrderedDict(
    {
        WarehouseAssetGroup.CATALOG: CATALOG_ASSETS,
        WarehouseAssetGroup.QUERY: QUERIES_ASSETS,
        WarehouseAssetGroup.VIEW_DDL: VIEWS_ASSETS,
        WarehouseAssetGroup.ROLE: (WarehouseAsset.USER,),
    },
)



from .collection import group_by
from .constants import OUTPUT_DIR
from .deprecate import deprecate_python
from .env import from_env
from .files import explode, search_files
from .formatter import to_string_array
from .json_stream_write import StreamableList
from .load import load_file
from .object import deep_serialize, getproperty
from .pager import (
    Pager,
    PagerLogger,
    PagerOnId,
    PagerOnIdLogger,
    PagerStopStrategy,
)
from .retry import RetryStrategy, retry
from .safe import SafeMode, safe_mode
from .store import AbstractStorage, LocalStorage
from .string import decode_when_bytes, string_to_tuple
from .time import (
    at_midnight,
    current_date,
    current_datetime,
    current_timestamp,
    past_date,
)
from .type import Callback, Getter, JsonType, SerializedAsset
from .uri import uri_encode
from .validation import validate_baseurl
from .write import (
    get_output_filename,
    get_summary_filename,
    get_summary_payload,
    write_errors_logs,
    write_json,
    write_summary,
)



# environment variable names
API_TOKEN = "CASTOR_DOMO_API_TOKEN"  # noqa: S105
BASE_URL = "CASTOR_DOMO_BASE_URL"
CLIENT_ID = "CASTOR_DOMO_CLIENT_ID"
CLOUD_ID = "CLOUD_ID"
DEVELOPER_TOKEN = "CASTOR_DOMO_DEVELOPER_TOKEN"  # noqa: S105



import io
import os

import pytest

from .formatter import (
    CsvFormatter,
    Formatter,
    JsonFormatter,
    _remove_unsupported_byte,
    to_string_array,
)


def test__to_string_array():
    assert to_string_array('["foo"]') == ["foo"]
    assert to_string_array('["foo", "bar"]') == ["foo", "bar"]
    assert to_string_array('["1", "2"]') == ["1", "2"]

    with pytest.raises(ValueError):
        to_string_array("")  # empty string
        to_string_array('"foo"')  # missing brackets
        to_string_array('["foo", 3]')  # all elements must be string


SAMPLE_KEYS = (
    "id",
    "first_name",
    "last_name",
    "email",
    "ip_address",
    "phone_number",
    "timezone",
    "has_subscribed",
)

SAMPLE_SIZE = 50

SAMPLE_FILE = "formatter_test"


def _test(formatter: Formatter) -> None:
    root = os.path.dirname(__file__)
    path = os.path.join(root, f"{SAMPLE_FILE}.{formatter.extension()}")
    with open(path, "r") as file:
        rows = list(formatter.deserialize(file))
        first = rows[0]

        assert len(rows) == SAMPLE_SIZE
        assert tuple(first.keys()) == SAMPLE_KEYS

        with io.StringIO() as f:
            # writing CSV to buffer
            formatter.serialize(f, rows)
            # reset the buffer position to the beginning before reading
            f.seek(0)
            # deserialize(serialize(x)) == x
            assert list(formatter.deserialize(f)) == rows


def test__csv_formatter():
    formatter = CsvFormatter()
    _test(formatter)


def test__json_formatter():
    formatter = JsonFormatter()
    _test(formatter)


def test__remove_unsupported_byte():
    row = [1, "foo", "bar\x00bie"]
    cleaned = _remove_unsupported_byte(row)
    assert cleaned == [1, "foo", "barbie"]



import pytest

from .validation import validate_baseurl


def test_validate_baseurl():
    assert (
        validate_baseurl("www.payfit.looker.eu")
        == "https://www.payfit.looker.eu"
    )

    assert (
        validate_baseurl("https://swile.looker.eu:443")
        == "https://swile.looker.eu:443"
    )

    assert (
        validate_baseurl("productboard.looker.eu:19999")
        == "https://productboard.looker.eu:19999"
    )

    assert (
        validate_baseurl("www.payfit.looker.eu/")
        == "https://www.payfit.looker.eu"
    )

    assert (
        validate_baseurl("productboard.looker.eu:19999/")
        == "https://productboard.looker.eu:19999"
    )

    # Idempotence
    url = "https://swile.looker.eu:443"
    assert validate_baseurl(url) == validate_baseurl(validate_baseurl(url))

    with pytest.raises(ValueError):
        validate_baseurl("")

    with pytest.raises(ValueError):
        validate_baseurl("ftp://tesla.looker.eu")

    with pytest.raises(ValueError):
        validate_baseurl("https://tesla.looker.eu:19999?query=params")

    with pytest.raises(ValueError):
        validate_baseurl("https://tesla.looker.eu:19999#fragment")



import logging
from argparse import ArgumentParser

from castor_extractor.visualization import powerbi  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    parser.add_argument("-t", "--tenant_id", help="PowerBi tenant ID")
    parser.add_argument("-c", "--client_id", help="PowerBi client ID")
    parser.add_argument("-s", "--secret", help="PowerBi password")
    parser.add_argument(
        "-sc",
        "--scopes",
        help="PowerBi scopes, optional",
        nargs="*",
    )
    parser.add_argument("-o", "--output", help="Directory to write to")

    args = parser.parse_args()
    powerbi.extract_all(
        tenant_id=args.tenant_id,
        client_id=args.client_id,
        secret=args.secret,
        scopes=args.scopes,
        output_directory=args.output,
    )



from uuid import UUID

from .constant import FileType
from .upload import _path


def test__path():
    source_id = UUID("399a8b22-3187-11ec-8d3d-0242ac130003")
    file_type = FileType.VIZ
    file_path = "filename"

    path = _path(source_id, file_type, file_path)
    assert path == f"visualization-{source_id}/{file_path}"
